## Copyright (C) 2010 Mark Borgerding <mark@borgerding.net>
##
## This program is free software; you can redistribute it and/or modify it under
## the terms of the GNU General Public License as published by the Free Software
## Foundation; either version 3 of the License, or (at your option) any later
## version.
##
## This program is distributed in the hope that it will be useful, but WITHOUT
## ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
## FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
## details.
##
## You should have received a copy of the GNU General Public License along with
## this program; if not, see <http://www.gnu.org/licenses/>.

## -*- texinfo -*-
## @deftypefn {Function File} {@var{outdata} =} helscandeintrlv (@var{data}, @var{nrows}, @var{ncols}, @var{Nshift})
## @var{nrows}-by-@var{ncols}.
## @seealso{helscandeintrlv}
## @end deftypefn

function outdata = helscandeintrlv (data, Nrows, Ncols, Nshift)

  outdata = helscanintrlv (data, Nrows, Ncols, Nrows - Nshift);

endfunction
